title = 'bigbucket'
description = 'Python3 client for Bigbucket'
url = 'https://github.com/adrianchifor/bigbucket-python'
version = '0.2.0'
author = 'Adrian Chifor'
author_email = 'me@adrianchifor.com'
license = 'GNU General Public License v3 (GPLv3)'



class RequestError(Exception):
    """Generic request error or 500 response from server"""
    pass


class ConnectionError(RequestError):
    """Connection refused to server"""
    pass


class Timeout(RequestError):
    """Request timeout"""
    pass


class TooManyRedirects(RequestError):
    """Request redirected too many times"""
    pass


class NotFound(RequestError):
    """Request response is 404"""
    pass



from __future__ import annotations

import requests

from bigbucket.auth import authenticate
from bigbucket.utils import parse_response, handle_exception


class Client(object):
    def __init__(self, address: str = "http://localhost:8080",
                 timeout: float = 30.0, gcp_auth: bool = False):
        self.address = address
        self.timeout = timeout
        self.gcp_auth = gcp_auth
        self.headers = None
        self.jwt = None

    @authenticate
    def get_tables(self) -> list:
        try:
            r = requests.get(f"{self.address}/api/table",
                             timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        return parse_response(r)["tables"]

    def table(self, table_name: str) -> Table:
        return Table(table_name, address=self.address, timeout=self.timeout,
                     gcp_auth=self.gcp_auth, headers=self.headers, jwt=self.jwt)


class Table(object):
    def __init__(self, table_name: str, **kwargs):
        self.table_name = table_name
        self.address = kwargs.get("address")
        self.timeout = kwargs.get("timeout")
        self.gcp_auth = kwargs.get("gcp_auth")
        self.headers = kwargs.get("headers")
        self.jwt = kwargs.get("jwt")

    @authenticate
    def delete_table(self):
        try:
            params = {"table": self.table_name}
            r = requests.delete(f"{self.address}/api/table", params=params,
                                timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        parse_response(r)

    @authenticate
    def list_columns(self) -> list:
        try:
            params = {"table": self.table_name}
            r = requests.get(f"{self.address}/api/column", params=params,
                             timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        return parse_response(r)["columns"]

    @authenticate
    def delete_column(self, column_name: str):
        try:
            params = {"table": self.table_name, "column": column_name}
            r = requests.delete(f"{self.address}/api/column", params=params,
                                timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        parse_response(r)

    @authenticate
    def count_rows(self, prefix: str = None) -> int:
        try:
            params = {"table": self.table_name}
            if prefix:
                params["prefix"] = prefix
            r = requests.get(f"{self.address}/api/row/count", params=params,
                             timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        return int(parse_response(r)["rowsCount"])

    @authenticate
    def read_row(self, key: str, columns: list = None) -> dict:
        try:
            params = {"table": self.table_name, "key": key}
            if columns:
                params["columns"] = ",".join(columns)
            r = requests.get(f"{self.address}/api/row", params=params,
                             timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        return parse_response(r)[key]

    @authenticate
    def read_rows(self, prefix: str = None, columns: list = None, limit: int = None) -> dict:
        try:
            params = {"table": self.table_name}
            if prefix:
                params["prefix"] = prefix
            if columns:
                params["columns"] = ",".join(columns)
            if limit:
                params["limit"] = limit
            r = requests.get(f"{self.address}/api/row", params=params,
                             timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        return parse_response(r)

    @authenticate
    def set_row(self, key: str, column_value_map: dict):
        try:
            params = {"table": self.table_name, "key": key}
            r = requests.post(f"{self.address}/api/row", params=params, json=column_value_map,
                              timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        parse_response(r)

    @authenticate
    def delete_row(self, key: str):
        try:
            params = {"table": self.table_name, "key": key}
            r = requests.delete(f"{self.address}/api/row", params=params,
                                timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        parse_response(r)

    @authenticate
    def delete_rows(self, prefix: str):
        try:
            params = {"table": self.table_name, "prefix": prefix}
            r = requests.delete(f"{self.address}/api/row", params=params,
                                timeout=self.timeout, headers=self.headers)
        except Exception as e:
            handle_exception(e)

        parse_response(r)



import requests
import base64
import time
import json

from functools import wraps


def authenticate(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.gcp_auth:
            self.headers, self.jwt = _get_gcp_auth_header(
                self.jwt, self.address)
        return func(self, *args, **kwargs)
    return wrapper


def _get_gcp_auth_header(jwt: str, address: str) -> (dict, str):
    """
    Constructs the auth header and returns it with a refreshed jwt token
    """
    if _jwt_expired(jwt):
        jwt = _get_gcp_auth_token(address)

    return {"Authorization": f"bearer {jwt}"}, jwt


def _get_gcp_auth_token(receiving_service_url: str) -> str:
    token_request_url = ("http://metadata/computeMetadata/v1/instance/service-accounts"
                         f"/default/identity?audience={receiving_service_url}")
    token_response = requests.get(token_request_url,
                                  headers={"Metadata-Flavor": "Google"})
    return token_response.content.decode("utf-8")


def _jwt_expired(jwt: str) -> bool:
    if not jwt:
        return True

    jwt_arr = jwt.split(".")
    jwt_payload = base64.b64decode(f"{jwt_arr[1]}==").decode()
    expiry = int(json.loads(jwt_payload)["exp"])
    # Add 30 seconds to allow time for requests
    now = int(time.time()) + 30

    return expiry <= now



import requests

from bigbucket.errors import ConnectionError, Timeout, TooManyRedirects, NotFound, RequestError


def parse_response(response):
    try:
        json_response = response.json()
    except ValueError:
        raise RequestError(
            f"{response.status_code}: Invalid response from server")

    if response.status_code == 404:
        raise NotFound(f"{response.status_code}: {json_response['error']}")
    if response.status_code != 200:
        raise RequestError(f"{response.status_code}: {json_response['error']}")

    return json_response


def handle_exception(exception):
    if isinstance(exception, requests.exceptions.ConnectionError):
        raise ConnectionError(exception)
    elif isinstance(exception, requests.exceptions.Timeout):
        raise Timeout(exception)
    elif isinstance(exception, requests.exceptions.TooManyRedirects):
        raise TooManyRedirects(exception)
    else:
        raise RequestError(exception)
